import fs from 'fs';
import os from 'os';
import path from 'path';
import { v4 as uuidv4 } from 'uuid';

import {
    getFileStoreMap,
    removeFromFileStoreMap,
    setFileStoreMap,
} from './redis.js';

/**
 * Retrieve a hash entry from Redis and ensure that the referenced files
 * still exist in at least one configured storage provider. If one copy is
 * missing it will try to restore it from the other provider (when possible).
 *
 * If the entry is completely invalid (no files found) it is removed from
 * the store and `null` is returned.
 *
 * The function also updates the timestamp of the entry so that active hashes
 * stay fresh in Redis.
 *
 * @param {object} context   – Azure Function context for logging
 * @param {string} hash      – The hash / key in the FileStoreMap
 * @param {StorageService} storageService – An initialised StorageService instance
 * @returns {object|null} The (possibly refreshed) entry or null when invalid
 */
export async function getValidHashEntry(context, hash, storageService) {
    if (!hash) return null;

    let entry = await getFileStoreMap(hash);
    if (!entry) return null;

    try {
        const primaryExists = entry?.url ? await storageService.fileExists(entry.url) : false;
        const gcsExists = entry?.gcs ? await storageService.fileExists(entry.gcs) : false;

        // If neither storage has the file, remove the entry and abort
        if (!primaryExists && !gcsExists) {
            await removeFromFileStoreMap(hash);
            return null;
        }

        // Restore missing GCS copy when primary exists
        if (primaryExists && !gcsExists) {
            try {
                entry = await storageService.ensureGCSUpload(context, entry);
            } catch (err) {
                context.log(`getValidHashEntry: failed to restore GCS copy – ${err}`);
            }
        }

        // Restore missing primary copy when GCS exists and a primary provider is configured
        if (!primaryExists && gcsExists && storageService.backupProvider?.isConfigured()) {
            let tempDir;
            let downloadedFile;
            try {
                tempDir = path.join(os.tmpdir(), `${uuidv4()}`);
                fs.mkdirSync(tempDir);
                downloadedFile = path.join(tempDir, path.basename(entry.gcs));

                // Download from GCS, then upload to primary storage
                await storageService.downloadFile(entry.gcs, downloadedFile);
                const res = await storageService.uploadFile(context, downloadedFile, hash);
                entry.url = res.url;
            } catch (err) {
                context.log(`getValidHashEntry: failed to restore primary copy – ${err}`);
            } finally {
                // Clean temp artefacts
                try {
                    if (downloadedFile && fs.existsSync(downloadedFile)) {
                        fs.unlinkSync(downloadedFile);
                    }
                    if (tempDir && fs.existsSync(tempDir)) {
                        fs.rmSync(tempDir, { recursive: true, force: true });
                    }
                } catch (_) {
                    /* noop */
                }
            }
        }

        // Update timestamp so the entry stays fresh
        await setFileStoreMap(hash, entry);
        return entry;
    } catch (err) {
        context.log(`getValidHashEntry: error during validation – ${err}`);
        await removeFromFileStoreMap(hash);
        return null;
    }
} 